"""Driver.py
"""

# Abstract Network Simulator ('anx')
#
# Copyright 2017 Vladimir Likic
# Copyright Commonwealth of Australia
#
# This software is licensed under the 3 Clause BSD License:
# https://opensource.org/licenses/BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=invalid-name

import simpy

from . import Globals
from . import Simulator
from . import ProgressBar
from . import TraceUtils


def run_sim(M, t, output_dir=None, bar=False, verbose=Globals.VERB_NO):
    """
    Runs network simulation based on the network model 'M'.
    """

    print(" [+] Initialising the simulation...")
    print(" [+] Total simulation time is {:.2f} time units".format(t))

    # initialise simpy environment
    env = simpy.Environment()

    # bind the network model 'M' to the SimPy simulation environment
    network = Simulator.setup_network(env, M, verbose=verbose)
    print(" [+] Simulations started...")

    # show progress bar
    if bar:
        print("\n [ Running progress bar ]")
        progress_bar = ProgressBar.setup(env, t)
        progress_bar.run()

    # run the simulation
    env.run(until=t)

    if bar:
        print("\n")

    print(" [+] Simulation completed")

    # print summary statistics
    TraceUtils.print_stats(network, verbose=verbose)

    # if 'output_dir' is defined, save node names, queue monitor,
    # and traces
    if output_dir is not None:
        TraceUtils.save_node_names(network, output_dir)
        TraceUtils.save_queue_mon(output_dir, network)
        # save generated, forwarded, received, discarded packets
        for trace in ['g', 'f', 'r', 'd']:
            TraceUtils.save_node_pkts(network, output_dir, trace)
